import json
from typing import Tuple, Dict

from petisco.events.event import Event

REQUEST_RESPONDED_UNWRAP_ERROR = (
    {
        "error": {
            "type": "RequestRespondedUnwrapError",
            "message": "Cannot unwrap http response from decorated function",
        }
    },
)

REQUEST_RESPONDED_VERSION = "1.2.0"


def is_flask_response(value):
    try:
        from flask import Response

        if isinstance(value, Response):
            return True
    except (RuntimeError, ImportError):
        pass
    return False


def get_content(value):
    if is_flask_response(value):
        return "flask response"
    else:
        return json.dumps(value)


class RequestResponded(Event):
    application: str
    controller: str
    is_success: bool
    http_response: Dict
    event_correlation_id: str
    elapsed_time: float
    additional_info: Dict[str, str]
    event_version: str = None

    def __init__(
        self,
        application: str,
        controller: str,
        is_success: bool,
        http_response: Tuple[Dict, int],
        correlation_id: str,
        elapsed_time: float,
        additional_info: Dict[str, str],
        version: str = REQUEST_RESPONDED_VERSION,
    ):
        self.application = application
        self.controller = controller
        self.is_success = is_success
        self.correlation_id = correlation_id
        self.elapsed_time = elapsed_time
        self.additional_info = additional_info
        self.set_http_response(http_response)
        self.version = version
        super().__init__()

    def set_http_response(self, http_response):
        _http_response = {
            "content": json.dumps(REQUEST_RESPONDED_UNWRAP_ERROR),
            "status_code": 500,
        }
        if isinstance(http_response, Tuple):
            _http_response["content"] = get_content(http_response[0])
            _http_response["status_code"] = http_response[1]
        else:
            _http_response["content"] = get_content(http_response)
            _http_response["status_code"] = http_response.status_code

        self.http_response = _http_response



from typing import Dict, Callable

from petisco.events.event import Event
from petisco.events.interface_event_manager import IEventManager


class NotImplementedEventManager(IEventManager):
    def __init__(self, subscribers: Dict[str, Callable] = None):
        super().__init__(subscribers)

    def info(self) -> Dict:
        return {"name": self.__class__.__name__}

    def unsubscribe_all(self):
        pass

    def publish(self, topic: str, event: Event):
        pass



from typing import Any

from dataclasses import dataclass
from dataclasses_json import dataclass_json


@dataclass_json
@dataclass
class LogMessage:
    layer: str = None
    operation: str = None
    correlation_id: str = None
    message: Any = None



import inspect

from typing import Dict, Callable, Any

from dataclasses import dataclass

from petisco.application.repository import Repository
from petisco.application.service import Service
from petisco.application.singleton import Singleton
from petisco.events.interface_event_manager import IEventManager
from petisco.logger.interface_logger import ILogger, INFO, ERROR
from petisco.logger.not_implemented_logger import NotImplementedLogger


@dataclass
class ApplicationConfig(metaclass=Singleton):
    name: str
    mode: str
    logger = ILogger
    services_provider: Callable[[], Dict[str, Service]]
    repositories_provider: Callable[[], Dict[str, Repository]]
    event_manager: IEventManager
    options: Dict[str, Any]
    info: Dict

    @staticmethod
    def get_instance():
        try:
            return ApplicationConfig()
        except:  # noqa E722
            raise ImportError(
                f"ApplicationConfig must be configured when the application starts. ApplicationConfig has been requested before its initial configuration. Incorrectly called by {inspect.stack()[1][3]}"
            )

    def __init__(
        self,
        app_name: str,
        mode: str,
        logger: ILogger = NotImplementedLogger(),
        config_dependencies: Callable = None,
        config_persistence: Callable = None,
        services_mode_mapper: Dict[str, Callable] = None,
        repositories_mode_mapper: Dict[str, Callable] = None,
        event_manager: IEventManager = None,
        options: Dict[str, Any] = None,
    ):
        """
        Parameters
        ----------
        app_name
            Application name
        mode
            DeploymentMode define the toy_app mode of execution. If you're mapping services and repositories, please
            check given mode is mapped in services_mode_mapper and repositories_mode_mapper
        logger
            Pre configured logger
        config_dependencies
            Callable function to configure dependencies (e.g configure credentials in order to connect with a thrid-party
            toy_app.
        config_persistence
            Callable function to configure toy_app persistence (e.g configure a database)
        services_mode_mapper
            A dictionary to map DeploymentMode with a service provider function. This is used as a dependency injector
        repositories_mode_mapper
            A dictionary to map DeploymentMode with a repository provider function. This is used as a dependency injector
        event_manager
            A IEventManager valid implementation
        options
            A dictionary with specific toy_app options
        """

        self.app_name = app_name
        self.mode = mode
        self.logger = logger

        self.logger.log(
            INFO, f"Deploying {self.app_name} application in {self.mode} mode"
        )

        if config_dependencies:
            config_dependencies()

        if config_persistence:
            config_persistence()

        self.info = {}
        if services_mode_mapper:
            if mode not in services_mode_mapper:
                error_message = f"Mode {mode} not found in services_mode_mapper ({services_mode_mapper})"
                self.logger.log(ERROR, error_message)
                raise NotImplementedError(error_message)
            self.services_provider = services_mode_mapper[mode]

            info_services = {}
            for key, service in self.services_provider().items():
                if hasattr(service, "info"):
                    info_services[key] = service.info()
                else:
                    raise TypeError(f"{key} must implement info")
            self.info["services"] = info_services

        if repositories_mode_mapper:
            if mode not in repositories_mode_mapper:
                error_message = f"Mode {mode} not found in repositories_mode_mapper ({repositories_mode_mapper})"
                self.logger.log(ERROR, error_message)
                raise NotImplementedError(error_message)
            self.repositories_provider = repositories_mode_mapper[mode]

            info_repositories = {}
            for key, repository in self.repositories_provider().items():
                if hasattr(repository, "info"):
                    info_repositories[key] = repository.info()
                else:
                    raise TypeError(f"{key} repository must implement info")
            self.info["repositories"] = info_repositories

        if event_manager:
            self.event_manager = event_manager
            if hasattr(event_manager, "info"):
                self.info["event_managers"] = event_manager.info()
            else:
                raise TypeError(f"Given event_manager must implement info")

        self.options = options

        self.logger.log(INFO, f"Info: {self.info}")
        self.logger.log(INFO, f"Options: {self.options}")



from abc import ABCMeta, abstractmethod

from dataclasses import dataclass
from dataclasses_json import dataclass_json


@dataclass_json
@dataclass
class IPersistenceConnector:

    __metaclass__ = ABCMeta

    @abstractmethod
    def execute(self):
        raise NotImplementedError



from typing import List

from petisco.events.interface_event_manager import IEventManager
from petisco.events.not_implemented_event_manager import NotImplementedEventManager


class EventConfig:
    def __init__(
        self,
        event_manager: IEventManager = NotImplementedEventManager(),
        event_topic: str = "petisco",
        event_additional_info: List[str] = None,
    ):
        """
        Parameters
        ----------
        event_manager
            A IEventManager implementation. Default NotImplementedEventManager
        event_topic
            Event Topic. By default petisco.
        event_additional_info
            A list of additional info to get from function argument and add to RequestResponded Event
        """
        self.event_manager = event_manager
        self.event_topic = event_topic
        self.event_additional_info = event_additional_info

    def get_additional_info(self, kwargs):
        if not self.event_additional_info:
            return None

        additional_info = {}
        for info_key in self.event_additional_info:
            info_value = kwargs.get(info_key)
            if info_value:
                additional_info[info_key] = info_value

        return additional_info



from typing import Tuple

from abc import ABCMeta, abstractmethod


class HttpError:

    __metaclass__ = ABCMeta

    @abstractmethod
    def __init__(self, message: str = "Unknown Error", code: int = 500):
        self.message = message
        self.code = code

    def __repr__(self):
        return f"{self.__class__.__name__}: {self.code} | {self.message}"

    def handle(self) -> Tuple[dict, int]:
        return (
            {"error": {"type": self.__class__.__name__, "message": self.message}},
            self.code,
        )



from meiga import Result, Error, Success, isFailure

from petisco.events.event import Event


class EventFromRedisMessageConversionError(Error):
    pass


def event_from_redis_message(message) -> Result[Event, Error]:
    try:
        json_data = message["data"].decode("utf-8")
        event = Event.from_json(json_data)
        return Success(event)
    except:  # noqa E722
        return isFailure(EventFromRedisMessageConversionError())



from typing import Dict, Callable, List
from abc import ABCMeta, abstractmethod

import deprecation

from petisco.events.event import Event
from petisco import __version__


class IEventManager:

    __metaclass__ = ABCMeta

    def __init__(self, subscribers: Dict[str, Callable]):
        self.subscribers = subscribers

    def __repr__(self):
        return f"EventManager: [subscribers: {self.subscribers}]"

    @abstractmethod
    def info(self) -> Dict:
        raise NotImplementedError

    @abstractmethod
    def unsubscribe_all(self):
        raise NotImplementedError

    @abstractmethod
    def publish(self, topic: str, event: Event):
        raise NotImplementedError

    def publish_list(self, topic: str, events: List[Event]):
        for event in events:
            self.publish(topic, event)

    @deprecation.deprecated(
        deprecated_in="0.14.4",
        removed_in="1.0.0",
        current_version=__version__,
        details="Use the publish function instead",
    )
    def send(self, topic: str, event: Event):
        self.publish(topic, event)



from functools import wraps

from meiga import Failure

from petisco.controller.tokens.jwt_config import JwtConfig
from petisco.controller.tokens.jwt_errors import InvalidTokenError


class JwtDecorator(object):
    def __init__(self, jwt_config: JwtConfig = None):
        self.jwt_config = jwt_config

    def __call__(self, func, *args, **kwargs):
        @wraps(func)
        def wrapper(*args, **kwargs):

            if not self.jwt_config:
                return func(*args, **kwargs)

            token_info = kwargs.get("token_info")

            if not token_info:
                return Failure(
                    InvalidTokenError(
                        message=f"This entry point expects a valid {self.jwt_config.token_type} Token"
                    )
                )

            client_id = token_info.get("client_id")
            token_type = token_info.get("token_type")
            user_id = token_info.get("user_id")

            if not user_id or user_id == "null":
                user_id = None

            if (
                token_type != self.jwt_config.token_type
                or (self.jwt_config.require_user and not user_id)
                or (not self.jwt_config.require_user and user_id)
            ):
                return Failure(
                    InvalidTokenError(
                        message=f"This entry point expects a valid {self.jwt_config.token_type} Token"
                    )
                )

            del kwargs["token_info"]
            if self.jwt_config.require_user:
                return func(client_id, user_id, *args, **kwargs)
            else:
                return func(client_id, *args, **kwargs)

        return wrapper


jwt = JwtDecorator



from datetime import datetime
from typing import Dict

from petisco.controller.correlation_id import CorrelationId
from petisco.events.event_id import EventId

import json


class Event:
    event_id: EventId = None
    event_name: str = None
    event_occurred_on: str = None
    event_version: str = None
    event_correlation_id: CorrelationId = None

    def __init__(self, dictionary=None):
        if dictionary:
            self.__dict__.update(dictionary)

        self.event_id = (
            EventId.generate(str(dictionary)) if not self.event_id else self.event_id
        )
        self.event_name = (
            self.__class__.__name__ if not self.event_name else self.event_name
        )
        self.event_occurred_on = (
            datetime.utcnow() if not self.event_occurred_on else self.event_occurred_on
        )

    def __repr__(self):
        return f"[{self.event_name}: {json.loads(self.to_json())}]"

    def __eq__(self, other):
        if (
            isinstance(other, self.__class__)
            or issubclass(other.__class__, self.__class__)
            or issubclass(self.__class__, other.__class__)
        ):
            return self.to_dict() == other.to_dict()
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def to_dict(self) -> Dict:
        return self.__dict__

    @staticmethod
    def from_dict(dictionary: Dict):
        if "event_id" in dictionary and isinstance(dictionary["event_id"], str):
            dictionary["event_id"] = EventId(dictionary["event_id"])
        if "event_occurred_on" in dictionary and isinstance(
            dictionary["event_occurred_on"], str
        ):
            dictionary["event_occurred_on"] = datetime.strptime(
                dictionary["event_occurred_on"], "%Y-%m-%d %H:%M:%S.%f"
            )
        return Event(dictionary)

    def to_json(self):
        return json.dumps(self.to_dict(), default=self._datetime_to_str)

    def _datetime_to_str(self, o):
        if isinstance(o, datetime):
            return o.__str__()

    @staticmethod
    def from_json(event_json):
        event_dict = json.loads(event_json)
        return Event.from_dict(event_dict)
